#!/usr/bin/perl -w
#
# fixsrc.pl - fixes epoc src tree to lower case
#
# (C) 2001 Mal Minhas
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
# 0.03   12. Oct 2002   alfredh   added some mmp file keywords
# 0.02   10. Oct 2002   alfredh   more improvements
# 0.01   16. Nov 2001   alfredh   original version
#
use strict;
use warnings;
use File::Spec;
use File::Basename;
use Getopt::Long;

# Usage:
# --overwrite|o, Overwrite destination files if they already exist
# --source|s, Source directory
# --target|t, Target directory

my $version = "0.03";
my $overwrite;
my $srcdir = $ARGV[0];
my $targetdir = $ARGV[1];
GetOptions('overwrite' => \$overwrite, 'source=s' => \$srcdir, 'target=s' => \$targetdir);

unless ($srcdir && $targetdir){
    die "Usage: fixsrc.pl $version [--overwrite] --source <srcdir> --target <targetdir>\n";
}
unless (-d $srcdir){
    die "Supplied <srcdir> is not a directory\n";
}
unless ($ENV{EPOCROOT} && -d $ENV{EPOCROOT})
	{
	die "Your EPOCROOT environment variable is unset or does not refer to a directory\n";
	}

# make the target directory with "chmod a=rwx"
mkdir $targetdir,0777;

# start process of moving directories over
my $retval=movedir($srcdir,$targetdir);
print "Returned OK\n";

# recursive open dir and merge of directory contents
sub movedir {
    my $srcdir = shift;
    my $targetdir = shift;
    my @srcdirs;

    unless (opendir(SRC,$srcdir)){
	warn "Couldn't open directory $srcdir:$!\n";
    }
    @srcdirs = readdir SRC;
    closedir (SRC);

    foreach my $i (@srcdirs){
	if (($i eq ".") || ($i eq "..")) { # skip '.' and '..'
	    next;
	}
	my $j = $i;

	$j =~ tr/[A-Z]/[a-z]/;  # force to lower case
	$i = "$srcdir/$i";
	$j = "$targetdir/$j";
	$j =~ s| |_|g; # check for space
	print "src: $i, tgt: $j\n";
	
	if (-f $i) { # file
	    if (-e $j && !$overwrite){
		print "$j already exists - ignoring $i\n";
	    }
	    else {
		system("cp",$i,$j);
		
		# now we need to downinch the target file
		# text files only - binary files are untouched
		if (-T $j){
		  unixifyfile($j);
		  downinch($j, $i);
		}
	    }
	} 
	elsif ( -d $i) { # directory
	    # if $j doesn't exist at target, need to make it...
	    if (! -e $j){
		print "Creating new directory $j...\n";
		mkdir $j,0777;
	    }
	    movedir($i,$j);
	} 
	else {
	    print STDERR "problems: $i\n";
	}
    }
}

#
# First check that we have a suitable candidate for downinching...
#
sub downinch {
    my $file = shift;
    my $filename = $file;
    my $oldfile = shift;
    my $fixcont = 0;

    #
    # handle .mmp files
    #
    if ($file =~ /\.(mmp|mmpi|mph)$/){
	my $nfile="$file.tmp";
	# open file and go through line at a time
	open(FILE, "<$file") or die "Error on opening $file $!\n";
	my $nline;
	open(NFILE,">$nfile") or die "Error on opening $nfile $!\n"; 
	while (my $line = <FILE>){
	    $line = makerelative($line, $filename, $oldfile);
	    if ($line =~ /^(\s*SOURCEPATH|\s*SOURCE|\s*TARGETTYPE|\s*TARGETPATH|\s*TARGET|\s*USERINCLUDE|\s*SYSTEMINCLUDE|\s*LIBRARY|\s*DEFFILE|\s*RESOURCE|\s*START.*BITMAP|\s*SYSTEMRESOURCE|\s*LANG|UID|\s*EPOCSTACKSIZE|\s*DOCUMENT|\s*START.*WINS|\s*END|\s*BASEADDRESS|\s*AIF)(.*)/i){
		# gets the lines with keywords in .mmp
		my $upper = uc($1); # switch up bogus lower case on keywords
		my $lower = lc($2);
		$nline = "$upper$lower\n";
	    }
	    elsif ($line =~ /^(\s*NOSTRICTDEF)/i){
		$nline = uc $1;
	    }
	    elsif ($line =~ /^(\s*UID|\s*MACRO)(.*)/i){
		# gets the lines with keywords in .mmp
		my $upper = uc($1); # switch up bogus lower case on keywords
		$nline = "$upper$2\n";
	    }
	    elsif ($line =~ /(.*)\.mmp/i){
                # gets the first comment of .mmp
		$nline = lc($line);
	    }
	    elsif ($line =~ /\#include/) {
	      $nline = lc $line;
	    }
	    elsif ($fixcont) {
	      $nline = lc $line;
	    }
	    else {
		# leave everything else alone
		$nline = $line;
	    }

	    if ($nline =~ /\\\s*$/) {
		$fixcont = 1;
	    }
	    else {
		$fixcont = 0;
	    }
	    $nline =~ s/\\/\//g; # global replace of \ with / in line.
	    if ($fixcont) { # Need to deal with line continuations
		$nline =~ s/\/\s*$/\\\n/;
	    }
	    print NFILE $nline;
	}
	close FILE;
	close NFILE;
	unlink $file;
	system("mv",$nfile,$filename);
	return;
    }
    #
    # handle bld.inf files
    #
    if ($file =~ /.*\.(inf|inc|mk)$/){
	# do whatever needs to be done for bld.inf files
	my $nfile="$file.tmp";
	# open file and go through line at a time
	open(INFFILE, "<$file") or die "Error on opening $file $!\n";
	my $nline;
	open(NINFFILE,">$nfile") or die "Error on opening $nfile $!\n"; 
	my $prjplatforms = 0;
	while (my $line = <INFFILE>){
	    $line = makerelative($line, $filename, $oldfile);
	    if ($line =~ /^\s*PRJ_PLATFORMS/i){
		$nline = $line;
		$prjplatforms = 1;
	    }
	    elsif ($line =~ /^\s*PRJ_/i){
		# leave the keywords alone
		$nline = $line;
	        $prjplatforms = 0;
	    }
	    elsif ($line =~ /^MAKEFILE/i){
		# leave the keywords alone
		$nline = $line;
	    }
	    elsif ($line =~ /^(\/\/)/){
		# leave the comments alone
		$nline = $line;
		$nline =~ s/BLD\.INF/bld\.inf/g;
		}
	    elsif ($line =~ /^(#\sif|#if)/){
		# leave the #ifdefs alone
		$nline = $line;
		}
	    elsif ($line =~ /^#include/o)
		{
		$nline = lc $line;
		}
	    elsif ($prjplatforms == 0){
		$nline = lc($line);
		$nline =~ m|\"(.*)\".*\"(.*)\"|;
		if ($1 && $2) {
#		  print "yes 1=$1 2=$2\n";
		  my $tmp1 = $1;
		  my $tmp2 = $2;
		  $tmp1 =~ s| |_|g;
		  $tmp2 =~ s| |_|g;
		  $nline = "$tmp1 $tmp2\n"; # very ugly :(
#		  print "nline: $nline\n";
		}

	      }
	    else {
		$nline = $line;
	    }
	    $nline =~ s/\\/\//g; # global replace of \ with / in line.
	    print NINFFILE $nline;
	}
	close INFFILE;
	close NINFFILE;
	unlink $file;
	system("mv",$nfile,$filename);
	return;
    }
    #
    # handle .?by files
    #
    if ($file =~ /\.?by$/){
	my $nfile="$file.tmp";
	# open file and go through line at a time
	open(INFFILE, "<$file") or die "Error on opening $file $!\n";
	my $nline;
	open(NINFFILE,">$nfile") or die "Error on opening $nfile $!\n"; 
	while (my $line = <INFFILE>){
	    if ($line =~ /^(file=|data=)/i){
	      $nline = $line; # do nothing for now
	    }
	    elsif ($line =~ /\#include/) {
	      $nline = lc $line;
	    }
	    else {
	      $nline = $line; # do nothing for now
	    }
	    $nline =~ s/\\/\//g; # global replace of \ with / in line.
	    print NINFFILE $nline;
	}
	close INFFILE;
	close NINFFILE;
	unlink $file;
	system("mv",$nfile,$filename);
	return;
    }

    # otherwise we (probably!) have a .cpp or .inl or .h file and need to downinch it.
    if ($file =~ /\.(cpp|c|h|inl|rh|hrh|rss|lag)$/){
      downinchfile($file)
    }

}

#
# Converts any absolute paths to relative paths in the given line
# Arguments:
# 1/ The absolute path to convert
# 2/ The location of the base file from which the relative path is measured
# 3/ The old base file, used to check that the relative path works  
#

sub makerelative
	{
	my $line = shift;
	my $file = shift;
	my $oldfile = shift;

	# This regexp says the following:
	# 1/ At the begining of a line or after a space
	# 2/ Look for something that looks a bit like a path
	# 3/ A path is a series of slashes followed by non-slashes etc.
	# 4/ If you find a path, try and convert it

	$line =~ s/(\A|\s)((?:(?:\/|\\)[^(?:\/|\\|\*)\s\r]+)+)/sprintf("$1%s", convertpath($2, $file, $oldfile))/ge;
	return $line;
	}

#
# Converts a path to a file from an absolute path to a relative path
#

sub convertpath
        {
        # Get the absolute path to the file
        my $original = shift;
	my $abspath = $original;
	$abspath =~ s/\\/\//g;

	# This the base path to the new bld.inf (etc) to which the relative path should refer
        my $base = shift;

	# The path to the old base file, used to check the relative path works
	my $oldbase = shift;

	# Ignore things referring to epoc32. EPOCROOT sorts these out
        return $original if $abspath =~ /^\/epoc32\//oi;

        # Generate the relative version of the path
        my $result = File::Spec->abs2rel("$ENV{EPOCROOT}$abspath", dirname $base);

	# Check we can find the old file using the relative path in case we got a false match from the regexp
	my $test = dirname($oldbase) . "/$result";
        $test =~ s/\.(def|frz)$/u$&/oi unless -e $test; # Bloomin def files don't have the same name
	print "Failed to convert path: $original to $result\n" unless -e $test;
	return $original unless -e $test;

        # Return the result
	$result = lc $result; 
	print "Converting Path: $original to $result\n";
        return $result;
        }

#
# lower case the following TEXT: #include <TEXT> or #include "TEXT"
#
sub downinchfile() {
    my $file = shift;
    my $filename = $file;

    my $nfile="$file.tmp";
    # open file and go through line at a time
    open(INFFILE, "<$file") or die "Error on opening $file $!\n";
    my $nline;
    open(NINFFILE,">$nfile") or die "Error on opening $nfile $!\n"; 
    while (my $line = <INFFILE>){
	if ($line =~ /^(\#include\s*)/){
	    $nline = lc($line);
	    $nline =~ s/\\/\//g; # global replace of \ with / in line.
	}
	else {
	    $nline = $line;
	}
	print NINFFILE $nline;
    }
    close INFFILE;
    close NINFFILE;
    unlink $file;
    system("mv",$nfile,$filename);
}


#
# Global search and replace of \r\n with \n in text files
#
sub unixifyfile() {
    my $file = shift;
    my $unixify;

    # no \n delimiters
    local $/=undef;
    #undef $/; # alias $INPUT_RECORD_SEPARATOR, $RS

    # open and slurp the file into $_
    open(FILE, $file) or die "Error: $!\n";
    $_ = <FILE>;
    close FILE;

    # do the regexp here.  suffix: globally and multiline
    $unixify = s/\r\n/\n/g;

    if(! m/\n$/) {
      print "adding missing newline for file '$file'\n";
      $_ = "$_\n";
      $unixify = 1;
    }

    if ($unixify){
	system("chmod", "a+w", $file);
	print "unixified $file\n";
	open FOUT, ">$file";
	print FOUT;
	close FOUT;
    }
}

